MODULE PZImage;

(* Last edited on 1999-08-06 21:58:52 by hcgl *)

IMPORT FileRd, FileWr, OSError, PGM, Wr, Rd, Thread;
IMPORT PZLR3Chain, PZIntChain;
FROM PZTypes IMPORT LONG, INT, NAT, BOOL;

TYPE
  Point = PZIntChain.Point;

PROCEDURE ReadPGM(fName: TEXT): REF T =
  VAR NX, NY, maxVal: NAT;
      aux : NAT;
  <* FATAL OSError.E, Rd.Failure, Thread.Alerted *>
  BEGIN 
    WITH
      rd = FileRd.Open(fName & ".pgm"),
      pgmRd = NEW(PGM.Reader).init(rd, NX, NY, maxVal),
      rM = NEW(REF T, NY, NX),
      M = rM^
    DO
     <* ASSERT maxVal <= 255 *>
     FOR y := NY-1 TO 0 BY -1  DO 
       FOR x := 0 TO NX-1 DO 
         pgmRd.get(aux);
         M[y,x]:= aux
      END
    END;  
    RETURN rM  
    END 
  END ReadPGM;


PROCEDURE WritePGM(fName: TEXT; READONLY M: T)=
  <* FATAL OSError.E, Wr.Failure, Thread.Alerted *>
  BEGIN
    WITH
      NY = NUMBER(M),
      NX = NUMBER(M[0]),
      wr = FileWr.Open(fName & ".pgm"),
      pgmWr = NEW(PGM.Writer).init(wr,NX, NY, 255, binary := TRUE)
    DO
      FOR y := NY-1 TO 0 BY -1 DO 
        FOR x := 0 TO NX-1 DO 
          pgmWr.put(M[y,x])
        END;
      END;
      pgmWr.finish();
      Wr.Close(wr)
    END
  END WritePGM;

PROCEDURE ExtractBoundary(
    READONLY M: T; 
    black: NAT;
    grey : NAT;
    side: Side;
  ): REF PZLR3Chain.T =
  VAR 
    x, y: NAT;
    Ok: BOOL := FALSE;
    length: NAT;
  BEGIN
    WITH
      NY = NUMBER(M),
      NX = NUMBER(M[0])
    DO
      y := 0; 
      WHILE y < NY AND NOT Ok  DO  
        x := 0; 
        WHILE x < NX AND NOT Ok DO 
           Ok := M[y,x] > black; 
           (* Ok := M[y,x] # 0; *)
           (* Ok := FLOAT(M[y,x], LONG) >= c + 0.05d0 * (255.0d0-c) *)
          IF NOT Ok THEN INC(x) END
        END;
        IF NOT Ok  THEN INC(y) END 
      END;
      <* ASSERT Ok *>
      length := 0;
      WITH
        rf = NEW(REF PZLR3Chain.T, NX*NY),
        fchain = rf^
      DO
        FollowBoundary(M, fchain, length, x, y, black, grey, side);
        RETURN PZLR3Chain.Trim(fchain, 0, length)
      END;
    END
  END ExtractBoundary; 


PROCEDURE FollowBoundary(
    READONLY M: T; 
    VAR f: PZLR3Chain.T;
    VAR  nSteps: NAT; 
    x, y: NAT;
    black: NAT;
    grey : NAT;
    side: Side := Side.Interpolated; 
  ) =
  (*
    Given the image "M" of an isolated piece,
    and a pixel "(x,y)" on the boundary of the piece
    (such that pixel "(x-1,y)" is outside),
    follows the piece's boundary counterclockwise.
    The points are stored in "f", and their 
    number is stored in "nSteps". *)

  VAR po : Point;  (* The "outside" foot *)
      pi : Point;  (* The "inside" foot *)
 
  PROCEDURE  Inside(x, y: NAT): BOOL =
    BEGIN
      RETURN M[y,x] > black
    END Inside;
 
  PROCEDURE Outside(x, y: NAT): BOOL =
    BEGIN
      RETURN M[y,x] <= black
    END Outside; 
 
  PROCEDURE Interpolate(xi, yi, xo, yo : NAT) : LONG =
    BEGIN 
      WITH 
        alpha = FLOAT(grey - M[yo,xo], LONG) /
                FLOAT(M[yi,xi] - M[yo,xo], LONG) 
      DO 
        RETURN MIN(2.0d0,MAX(-1.0d0,alpha))
      END
    END Interpolate;
 
  PROCEDURE StoreTail( xi, yi, xo, yo : INT ) =
    BEGIN
      CASE side OF 
        Side.Interpolated =>
          WITH 
            alpha = Interpolate(xi, yi, xo, yo)
          DO 
            f[nSteps,0] := (1.0d0 - alpha)*FLOAT(xo,LONG)+alpha*FLOAT(xi, LONG);
            f[nSteps,1] := (1.0d0 - alpha)*FLOAT(yo,LONG)+alpha*FLOAT(yi, LONG);
          END
      | Side.Inner =>  
          f[nSteps,0] := FLOAT(xi, LONG);
          f[nSteps,1] := FLOAT(yi, LONG);
      | Side.Outer =>
          f[nSteps,0] := FLOAT(xo, LONG);
          f[nSteps,1] := FLOAT(yo, LONG);
      END;
      f[nSteps,2] := 0.0d0;
      INC(nSteps)   
    END StoreTail; 
  
  BEGIN

    pi := Point{x, y};
    po := Point{x, y-1}; 
     
    REPEAT
      WITH
        xo = po[0],
        yo = po[1],
        xi = pi[0],
        yi = pi[1]
      DO
        StoreTail(xi,yi,xo,yo);
        IF yo = yi-1 THEN                     
          IF Inside(xi+1, yi) THEN 
            IF Outside(xo+1, yo) THEN 
              pi[0] := xi+1; po[0] := xo+1;
            ELSE 
              pi[0] := xi+1; pi[1] := yi-1;
            END;
          ELSE 
            po[0] := xo+1; po[1] := yo+1; 
          END;
        ELSIF yo = yi+1  THEN                     
          IF Inside(xi-1, yi) THEN 
            IF Outside(xo-1, yo) THEN 
              pi[0] := xi-1; po[0] := xo-1;
            ELSE 
              pi[0] := xi-1; pi[1] := yi+1;
            END;
          ELSE 
            po[0] := xo-1; po[1] := yo-1 
          END;                
        ELSIF xo = xi+1 THEN                     
          IF Inside(xi, yi+1) THEN 
            IF Outside(xo, yo+1) THEN 
              pi[1] := yi+1; po[1] := yo+1;
            ELSE 
              pi[0] := xi+1; pi[1]:= yi+1;
            END;
          ELSE 
            po[0] := xo-1; po[1] := yo+1
          END; 
        ELSIF xo = xi-1  THEN                     
          IF Inside(xi, yi-1) THEN 
            IF Outside(xo, yo-1) THEN 
              pi[1] := yi-1; po[1] := yo-1;
            ELSE 
              pi[0] := xi-1; pi[1] := yi-1;
            END;
          ELSE 
            po[0] := xo+1; po[1] := yo-1
          END;                  
        END; 
     END;
    UNTIL ( pi[0] = x AND pi[1] = y ) AND ( po[0] = x AND po[1] = y-1 ) ;
  END FollowBoundary;

 
BEGIN
END PZImage.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
