MODULE PZLRChain;
(* Last edited on 1999-11-15 00:05:10 by hcgl *)

IMPORT NGet, NPut, FGet, FPut, FileFmt, Rd, Wr, Thread, FileRd, OSError, Fmt;
IMPORT PZProc, PZMatch, PZSegment, PZCandidate;

FROM Math IMPORT sqrt;
FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS, INT, NAT, BOOL, BOOLS;

PROCEDURE Trim(READONLY c: T; start, length: NAT): REF T =
  BEGIN
    WITH rx = NEW(REF T, length), x = rx^ DO
      DoTrim(c, start, length, x);
      RETURN rx
    END;
  END Trim;

PROCEDURE DoTrim(READONLY c: T; start, length: NAT; VAR x: T) =
  BEGIN
    WITH
      N = NUMBER(c),
      ini = start MOD N,
      fin = ini + length - 1
    DO
      IF fin < N  THEN
        x := SUBARRAY(c, ini, length)
      ELSE
        SUBARRAY(x, 0, N-ini) := SUBARRAY(c, ini, N-ini);
        SUBARRAY(x, N-ini, (fin+1) MOD N):= SUBARRAY(c, 0, (fin+1) MOD N);
      END;
    END;
  END DoTrim;

PROCEDURE Reverse(VAR c: T) =
  VAR aux: LONG;
  BEGIN
    WITH N = NUMBER(c), L = N-1 DO
      FOR i := 0 TO (N DIV 2)-1  DO
        aux := c[i]; c[i] := c[L-i]; c[L-i] := aux
      END;
    END;
  END Reverse;

PROCEDURE Complement(VAR c: T) =
  BEGIN
    WITH N = NUMBER(c) DO
      FOR i := 0 TO N-1 DO c[i] := -c[i] END
    END;
  END Complement;

PROCEDURE ExtractSegment(READONLY c: T; s: PZSegment.T; curvature: BOOL): REF T =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    WITH rx = NEW(REF T, s.ns), x = rx^ DO
      DoExtractSegment(c, s, curvature, x);
      RETURN rx
    END
  END ExtractSegment;

PROCEDURE DoExtractSegment(READONLY c: T; s: PZSegment.T; curvature: BOOL; VAR x: T) =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    DoTrim(c, s.ini, s.ns, x);
    IF s.rev THEN
      Reverse(x);
      IF curvature THEN Complement(x) END;
    END;
  END DoExtractSegment;

CONST
  FileVersion  = "97-10-29";

PROCEDURE Read(rd: Rd.T; headerOnly: BOOL): ReadData =

  PROCEDURE ToLong(n: INT): LONG =
    BEGIN
      <* ASSERT n > FIRST(INT) *>
      <* ASSERT n < LAST(INT) *>
      RETURN FLOAT(n, LONG)
    END ToLong;

  VAR d : ReadData;
  BEGIN
    FileFmt.ReadHeader(rd, "PZLRChain.T", FileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    d.stride := NGet.LongReal(rd, "stride");
    FGet.Skip(rd);
    d.unit := NGet.LongReal(rd, "unit");
    FGet.Skip(rd);
    d.samples := NGet.Int(rd, "samples"); FGet.EOL(rd);
    WITH
      N = d.samples,
      unit = d.unit
    DO
      IF headerOnly THEN
        d.c := NIL
      ELSE
        d.c := NEW(REF T, N);
        WITH c = d.c^ DO
          FOR j := 0 TO N-1 DO
            FGet.Skip(rd);
            c[j] := ToLong(FGet.Int(rd)) * unit
          END
        END;
        FileFmt.ReadFooter(rd, "PZLRChain.T")
      END;
      RETURN d
    END
  END Read;

PROCEDURE Write(wr :Wr.T; READONLY cmt : TEXT;  READONLY c: T; stride: LONG; unit: LONG) =

  PROCEDURE ToInt(x: LONG): INT =
    BEGIN
      <* ASSERT x > FLOAT(FIRST(INT),LONG) *>
      <* ASSERT x < FLOAT(LAST(INT), LONG) *>
      RETURN ROUND(x)
    END ToInt;

  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    FileFmt.WriteHeader(wr, "PZLRChain.T", FileVersion);
    FileFmt.WriteComment(wr, cmt, '|');
    NPut.LongReal(wr, "stride", stride); FPut.EOL(wr);
    NPut.LongReal(wr, "unit", unit); FPut.EOL(wr);
    NPut.Int(wr, "samples", NUMBER(c)); FPut.EOL(wr);
    FOR i := 0 TO LAST(c)  DO
      FPut.Int(wr, ToInt(c[i]/unit)); FPut.EOL(wr)
    END;
    FileFmt.WriteFooter(wr, "PZLRChain.T");
    Wr.Flush(wr)
  END Write;

PROCEDURE ReadAll(
    prefix: TEXT;
    band: NAT;
    extension: TEXT;
    READONLY sel: BOOLS;
    headerOnly: BOOL;
    dir: TEXT := ".";
  ): ReadAllData =
  VAR unitMin: LONG := LAST(LONG);
      unitMax: LONG := FIRST(LONG);
      cmt: TEXT := ""; 
  CONST NoData = ReadData{
    cmt := "NOT READ", samples := 0, c := NIL, stride := 0.0d0, unit := 0.0d0
  };
  <* FATAL Wr.Failure, Thread.Alerted, Rd.Failure *>
  BEGIN
    Wr.PutText(stderr, "reading numeric chains:\n");
    WITH
      rr = NEW(REF ARRAY OF ReadData, NUMBER(sel)), r = rr^
    DO
      FOR k := 0 TO LAST(sel) DO
        IF sel[k] THEN
          Wr.PutText(stderr, "  " & Fmt.Pad(Fmt.Int(k), 4, '0') & ": ");
          WITH
            chainTag  = Fmt.Pad(Fmt.Int(k), 4, '0'),
            bandTag = Fmt.Pad(Fmt.Int(band), 3, '0'),
            fileName = prefix & bandTag & extension,
            fullName = dir & "/" & chainTag &"/"& fileName,
            rd = OpenRead(fullName)
          DO
            IF rd # NIL THEN
              WITH data = Read(rd, headerOnly := headerOnly) DO
                Wr.PutText(stderr, "read from " & fullName & "\n");
                r[k] := data;
                <* ASSERT data.unit > 0.0d0 *>
                IF unitMin > unitMax THEN 
                  cmt := data.cmt
                END;
                unitMin := MIN(unitMin, data.unit);
                unitMax := MAX(unitMax, data.unit);
              END;
              Rd.Close(rd)
            ELSE
              r[k] := NoData;
              Wr.PutText(stderr, "file " & fullName & " not found.\n")
            END
          END
        ELSE
          r[k] := NoData;
        END
      END;
      RETURN ReadAllData {
        chain := rr,
        unitMin := unitMin,
        unitMax := unitMax,
        cmt := 
          "Chains read from " & 
          dir & "/NNNN/" & prefix & "BBB" & extension & "\n" &
          cmt
      }
    END
  END ReadAll;

PROCEDURE OpenRead(name: TEXT): Rd.T =
  BEGIN
    TRY
      RETURN FileRd.Open(name)
    EXCEPT
      OSError.E => RETURN NIL
    END
  END OpenRead;

PROCEDURE FindBestMatch(
    READONLY a: T;
    READONLY b: T;
    step: LONG;
    maxDistSqr: LONG;
    skipDistSqr: LONG;
    removeUnmatchedEnds: BOOL;
    VAR (*OUT*) mt: REF PZMatch.T;
    VAR (*OUT*) avgDisc: LONG;
    VAR (*OUT*) length: LONG;
    VAR (*OUT*) matchedLength: LONG;
    VAR (*WORK*) rCost: REF PZMatch.CostMatrix;
  ) =

  TYPE Pair = PZMatch.Pair;

  PROCEDURE StepDistSqr(READONLY p, q: Pair): LONG =
    (*
      Computes the mean squared chain distance for one matching step "p -> q": *)
    BEGIN
      WITH
        dp = a[p[0]] - b[p[1]],
        dq = a[q[0]] - b[q[1]],
        mdSqr = 0.5d0 * (dp*dp + dq*dq)
      DO
        RETURN MIN(mdSqr, maxDistSqr)
      END
    END StepDistSqr;

  PROCEDURE StepIntgDiscSqr(READONLY p, q: Pair): LONG =
    (*
      Computes the integrated discrepancy squared
      for one matching step "p -> q": *)
    BEGIN
      WITH
        d2 = StepDistSqr(p, q)
      DO
        IF p[0] = q[0] OR p[1] = q[1] THEN
          RETURN step * (d2 + skipDistSqr)/2.0d0
        ELSE
          RETURN step * d2
        END
      END
    END StepIntgDiscSqr;

  PROCEDURE StepIsMatched(READONLY p, q: Pair): BOOL =
    (*
      A "PZMatch.StepPredicate" that returns TRUE if the step "p->q"
      has average chain distance less than "maxDist". *)
    BEGIN
      RETURN StepDistSqr(p, q) < maxDistSqr
    END StepIsMatched;

  VAR intgDiscSqr: LONG;
  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b)
    DO
      PZMatch.FindBest(
        NA, NB,
        stepCostFn := StepIntgDiscSqr,
        (*OUT*) mt := mt, totCost := intgDiscSqr,
        (*WORK*) rCost := rCost
      );
      IF removeUnmatchedEnds THEN
        mt := PZMatch.RemoveUnmatchedEnds(mt^, StepIsMatched);
        intgDiscSqr := PZMatch.Sum(mt^, stepFn := StepIntgDiscSqr)
      END;
      <* ASSERT NUMBER(mt^) > 0 *>
      WITH
        m = NUMBER(mt^),
        na = mt[m-1][0] - mt[0][0],
        nb = mt[m-1][1] - mt[0][1]
      DO
        length := step * FLOAT(na + nb, LONG)/2.0d0;
        IF length <= 0.0d0 THEN
          avgDisc := sqrt(maxDistSqr)
        ELSE
          avgDisc := sqrt(intgDiscSqr/length)
        END
      END;
      WITH msc = PZMatch.MatchedStepCount(mt^, StepIsMatched) DO
        matchedLength := step * FLOAT(msc, LONG)/2.0d0
      END
    END;
  END FindBestMatch;

PROCEDURE RefineMatch(
    READONLY a: T;
    READONLY b: T;
    step: LONG;
    READONLY mtOld: PZMatch.T;
    maxAdjust: NAT;
    critDistSqr: LONG;
    maxDistSqr: LONG;
    skipDistSqr: LONG;
    VAR (*OUT*) mt: REF PZMatch.T;
    VAR (*OUT*) mismatch: LONG;
    VAR (*OUT*) length: LONG;
    VAR (*OUT*) matchedLength: LONG;
    VAR (*OUT*) minAvgDisc: LONGS;
    VAR (*WORK*) rCost: REF PZMatch.CostMatrix;
  ) =

  TYPE Pair = PZMatch.Pair;

  PROCEDURE StepDistSqr (READONLY p, q: Pair): LONG =
    (*
      The mean square distance between the chain steps
      "a[p[0]]->a[q[0]]" and "b[p[1]]->b[q[1]]". *)
    BEGIN
      WITH
        dp = a[p[0]] - b[p[1]],
        dq = a[q[0]] - b[q[1]],
        mdSqr = 0.5d0 * (dp*dp + dq*dq)
      DO
        RETURN MIN(mdSqr, maxDistSqr)
      END
    END StepDistSqr;

  PROCEDURE StepMismatch (READONLY p, q: Pair): LONG =
    (*
      A "PZMatch.StepFunction" whose value is the mismatch
      of the step "p -> q", i.e. the integrated square
      distance between the chain steps "a[p[0]]->a[q[0]]"
      and "b[p[1]]->b[q[1]]", plus the skip distance squared,
      minus the critical distance squared,
      all times the step length. *)
    BEGIN
      WITH
        d2 = StepDistSqr(p, q) - critDistSqr
      DO
        IF p[0] = q[0] OR p[1] = q[1] THEN
          RETURN step * (d2 + skipDistSqr)/2.0d0
        ELSE
          RETURN step * d2
        END
      END
    END StepMismatch;

  PROCEDURE StepIsMatched(READONLY p, q: Pair): BOOL =
    (*
      A "PZMatch.StepPredicate" that returns TRUE if the step "p->q"
      has average chain distance less than "maxDist". *)
    BEGIN
      RETURN StepDistSqr(p, q) < maxDistSqr
    END StepIsMatched;

  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b)
    DO
      <* ASSERT NA > 0 *>
      <* ASSERT NB > 0 *>
      PZMatch.Refine(
        NA := NA, NB := NB,
        stepCostFn := StepMismatch,
        mtOld := mtOld,
        maxAdjust := maxAdjust,
        (*OUT*) mt := mt,
        (*OUT*) totCost := mismatch,
        (*OUT*) minTotCost := minAvgDisc,
        (*WORK*) rCost := rCost
      );
      <* ASSERT  NUMBER(mt^) > 0 *>
      WITH
        mt = mt^,
        m = NUMBER(mt),
        stepCount = (mt[m-1][0] - mt[0][0]) + (mt[m-1][1] - mt[0][1]),
        matchedStepCount = PZMatch.MatchedStepCount(mt, StepIsMatched)
      DO
        length := FLOAT(stepCount, LONG) * step / 2.0d0;
        matchedLength := step * FLOAT(matchedStepCount, LONG)/2.0d0
      END;
      (* Compute the minimum average distance for each matching length "K". *)
      (* Note that "PZMatch.Refine" returns total mismatch, not just "intgDiscSqr" *)
      <* ASSERT  minAvgDisc[0] = 0.0d0 *>
      FOR K := 1 TO LAST(minAvgDisc) DO
        WITH Lk = step*FLOAT(K, LONG)/2.0d0 DO
          minAvgDisc[K] := sqrt(MAX(0.0d0, minAvgDisc[K]/Lk + critDistSqr))
        END
      END
    END;
  END RefineMatch;

PROCEDURE Delta(READONLY c: T; stride: LONG; ini, fin: INT): LONG =
  BEGIN
    WITH
      m = NUMBER(c),
      ri = c[ini MOD m] + stride * FLOAT(ini DIV m, LONG),
      rf = c[fin MOD m] + stride * FLOAT(fin DIV m, LONG)
    DO
      RETURN rf - ri
    END
  END Delta;

PROCEDURE ValueFromArg(READONLY c: T; stride: LONG; x: LONG): LONG =
  BEGIN
    WITH
      n = NUMBER(c),
      ka = FLOOR(x),
      kb = ka + 1,
      dx = x - FLOAT(ka, LONG),
      ya = c[ka MOD n] + stride * FLOAT(ka DIV n, LONG),
      yb = c[kb MOD n] + stride * FLOAT(kb DIV n, LONG)
    DO
      RETURN  ya + dx * (yb - ya)
    END
  END ValueFromArg;

PROCEDURE ValuesFromArgs(
    READONLY c: T;
    stride: LONG;
    READONLY x: LONGS;
    VAR y: LONGS;
  ) =
  BEGIN
    WITH
      n = NUMBER(c),
      m = NUMBER(x)
    DO
      <* ASSERT NUMBER(y) = m *>
      FOR i := 0 TO m-1 DO
        WITH
          ka = FLOOR(x[i]),
          kb = ka + 1,
          dx = x[i] - FLOAT(ka, LONG),
          ya = c[ka MOD n] + stride * FLOAT(ka DIV n, LONG),
          yb = c[kb MOD n] + stride * FLOAT(kb DIV n, LONG)
        DO
          y[i] := ya + dx * (yb - ya)
        END
      END;
    END
  END ValuesFromArgs;

PROCEDURE ArgFromValue(READONLY c: T; stride: LONG; y: LONG): LONG =
  VAR i, j, q: INT;
  BEGIN
    WITH
      n = NUMBER(c)
    DO
      (* Reduce "y" to the range "[c[0] _ c[0] + stride)" *)
      <* ASSERT stride > 0.0d0 *>
      q := FLOOR((y-c[0])/stride);
      y := y - FLOAT(q, LONG) * stride;
      WHILE y >= c[0] + stride DO INC(q); y := y - stride END;
      (* Binary search for the interval "[c[i] _ c{i+1])" that contains "y": *)
      <* ASSERT c[0] <= y AND y < c[0] + stride *>
      i:= 0; j := n;
      WHILE i < j-1 DO
        WITH
          m = (i+j) DIV 2
        DO
          <* ASSERT m < n *>
          IF y < c[m] THEN j := m ELSE i := m END;
        END
      END;
      (* Interpolation: *)
      WITH
        fi = FLOAT(i, LONG) + FLOAT(q*n, LONG),
        ci = c[i],
        cj = c[(i+1) MOD n] + stride * FLOAT((i+1) DIV n, LONG)
      DO
        <* ASSERT y >= ci AND y <= cj *>
        IF ci = cj THEN
          RETURN fi + 0.5d0
        ELSE
          RETURN fi + (y-ci)/(cj-ci)
        END
      END
    END;
  END ArgFromValue;

PROCEDURE ArgsFromValues(
    READONLY c: T;
    stride: LONG;
    READONLY y: LONGS;
    VAR x: LONGS;
  ) =
  VAR ia, ib: NAT;
      k: INT;
      yy: LONG;
  BEGIN
    WITH
      nc = NUMBER(c),
      ny = NUMBER(y),
      cLo = c[0],
      cHi = cLo + stride
    DO
      <* ASSERT stride > 0.0d0 *>
      <* ASSERT NUMBER(x) = ny *>
      ia := 0; ib := 1;
      FOR j := 0 TO ny-1 DO
        (* Invariant: "ia IN [0..np-1]", "ib = ia+1" *)
        (* Reduce "y[j]" to the range "[c[0] _ c[0] + stride)" *)
        yy := y[j];
        PZProc.ReduceToPeriod(yy, cLo, cHi, k);
        (* Sequential search for the interval "[c[ia] _ c[ib])" that contains "yy": *)
        <* ASSERT cLo <= yy AND yy < cHi *>
        WHILE ib < nc AND yy > c[ib] DO ia := ib; INC(ib) END;
        WHILE yy < c[ia] DO ib := ia; DEC(ia) END;
        (* Interpolation: *)
        WITH
          fi = FLOAT(ia, LONG) + FLOAT(k*nc, LONG),
          ci = c[ia],
          cj = c[ib MOD nc] + stride * FLOAT(ib DIV nc, LONG)
        DO
          <* ASSERT yy >= ci AND yy <= cj *>
          IF ci = cj THEN
            x[j] := fi + 0.5d0
          ELSE
            x[j] := fi + (yy-ci)/(cj-ci)
          END
        END
      END
    END;
  END ArgsFromValues;

PROCEDURE ValueFromValue(
    READONLY a: T; aStride: LONG;
    READONLY b: T; bStride: LONG;
    ya: LONG;
  ): LONG =
  BEGIN
    WITH x = ArgFromValue(a, aStride, ya) DO
      RETURN ValueFromArg(b, bStride, x);
    END;
  END ValueFromValue;

PROCEDURE ArgFromArg(
    READONLY a: T; aStride: LONG;
    READONLY b: T; bStride: LONG;
    xa: LONG;
  ): LONG =
  BEGIN
    WITH y = ValueFromArg(a, aStride, xa) DO
      RETURN ArgFromValue(b, bStride, y);
    END;
  END ArgFromArg;

PROCEDURE MapSegment(
    READONLY sOld: PZSegment.T;
    READONLY tOld: T;
    READONLY tNew: T;
    stride: LONG;
  ): PZSegment.T =
  BEGIN
    WITH
      cvx = sOld.cvx,

      iniOld = sOld.ini,
      finOld = iniOld + sOld.ns - 1, (* DON'T take MOD! *)

      labelIni = ValueFromArg(tOld,stride,FLOAT(iniOld,LONG)),
      labelFin = ValueFromArg(tOld,stride,FLOAT(finOld,LONG)),

      totNew = NUMBER(tNew),

      iniNew = ROUND(ArgFromValue(tNew, stride, labelIni)),
      finNew = ROUND(ArgFromValue(tNew, stride, labelFin))
    DO
      <* ASSERT iniNew <= finNew *>
      WITH
        mapSeg =  PZSegment.T{
        cvx := cvx,
        tot := totNew,
        ini := iniNew MOD totNew,
        ns := MIN(finNew - iniNew + 1, totNew + 1),
        rev := sOld.rev
      }
      DO
        RETURN mapSeg
      END
    END
  END MapSegment;

PROCEDURE MapMatch(
    READONLY mOld: PZMatch.T;
    READONLY sOld: PZSegment.Pair;
    READONLY t0Old, t1Old: T;
    READONLY sNew: PZSegment.Pair;
    READONLY t0New, t1New: T;
    stride: ARRAY [0..1] OF LONG;
  ): REF PZMatch.T =

  PROCEDURE AllocInitialMatch(): REF PZMatch.T =
    (* Allocates initial storage for the new matching *)
    BEGIN
      WITH
        nOld = NUMBER(t0Old) + NUMBER(t1Old),
        nNew = NUMBER(t0New) + NUMBER(t1New),
        mNewSizeEst = MAX(NUMBER(mOld) * nNew DIV nOld, 1)
      DO
        RETURN NEW(REF PZMatch.T, mNewSizeEst)
      END;
    END AllocInitialMatch;

  PROCEDURE MapAbsPair(pOld: PZMatch.Pair): PZMatch.Pair =
    (*
      Maps a pair of ABSOLUTE sample indices from the old
      curves to the new curves. *)
    BEGIN
      WITH
        label0 = ValueFromArg(t0Old,stride[0],FLOAT(pOld[0],LONG)),
        pNew0 = ROUND(ArgFromValue(t0New, stride[0], label0)),

        label1 = ValueFromArg(t1Old,stride[1],FLOAT(pOld[1],LONG)),
        pNew1 = ROUND(ArgFromValue(t1New, stride[1], label1))
      DO
        RETURN PZMatch.Pair{pNew0, pNew1}
      END
    END MapAbsPair;

  PROCEDURE EnsureSpace(VAR mR: REF PZMatch.T; k: NAT) =
    BEGIN
      IF mR = NIL THEN
        mR := NEW(REF PZMatch.T, k)
      ELSIF k > LAST(mR^) THEN
        WITH n = NUMBER(mR^), mT = NEW(REF PZMatch.T, MAX(k+1, 2*n)) DO
          SUBARRAY(mT^,0,n)  := mR^; mR := mT
        END
      END
    END EnsureSpace;

  VAR mNewR: REF PZMatch.T := AllocInitialMatch();  (* The new matching *)
      kNew: NAT; (* Number of new pairs in "mNewR^ *)
      stop: BOOL;  (* TRUE after first interruption. *)
  BEGIN
    kNew := 0;
    stop := FALSE;
    FOR kOld := 0 TO LAST(mOld) DO
      WITH
        rOld = mOld[kOld],                      (* Old pair relative to "sOld" *)
        pOld = PZMatch.AbsPairClip(rOld, sOld), (* Old pair, absolute *)
        pNew = MapAbsPair(pOld),                (* New pair, absolute *)
        rNew = PZMatch.RelPairClip(pNew, sNew)  (* New pair, relative to "sNew" *)
      DO
        IF rNew[0] = LAST(NAT) OR rNew[1] = LAST(NAT) THEN
          (* Pair is outside "sNew": *)
          IF kNew > 0 THEN stop := TRUE END
        ELSE
          IF stop THEN RETURN NIL END;
          (* Pair lies inside "sNew" *)
          IF kNew = 0 THEN
            (* Firstpair of new matching, just store it: *)
            EnsureSpace(mNewR, kNew);
            mNewR^[kNew] := rNew;
            INC(kNew)
          ELSE
            (* Interpolate, Bresenham style, as needed: *)
            VAR
              aNew := mNewR^[kNew-1];
              p := aNew;
            BEGIN
              <* ASSERT aNew[0] <= rNew[0] *>
              <* ASSERT aNew[1] <= rNew[1] *>
              WITH
                d0 = rNew[0] - aNew[0],
                d1 = rNew[1] - aNew[1],
                d = MAX(d0, d1)
              DO
                FOR ib := 1 TO d DO
                  IF d0 > d1 THEN
                    p[0] := aNew[0] + ib;
                    p[1] := aNew[1] + (ib*d1) DIV d
                  ELSE
                    p[0] := aNew[0] + (ib*d0) DIV d;
                    p[1] := aNew[1] + ib
                  END;
                  EnsureSpace(mNewR, kNew);
                  mNewR^[kNew] := p;
                  INC(kNew)
                END;
              END
            END
          END
        END;
      END
    END;
    IF kNew = 0 THEN RETURN NIL END;
    WITH mR = NEW(REF PZMatch.T, kNew) DO
      mR^ := SUBARRAY(mNewR^, 0, kNew);
      RETURN mR
    END
  END MapMatch;

PROCEDURE MapPackedMatch(
    pm: REF PZMatch.PackedT;
    READONLY sOld: PZSegment.Pair;
    READONLY t0Old, t1Old: T;
    READONLY sNew: PZSegment.Pair;
    READONLY t0New, t1New: T;
    stride: ARRAY [0..1] OF LONG;
  ): REF PZMatch.PackedT =
  BEGIN
    IF pm = NIL THEN
      RETURN NIL
    ELSE
      WITH
        mOld = PZMatch.Unpack(pm^),
        mNew = MapMatch(mOld^, sOld, t0Old, t1Old, sNew, t0New, t1New, stride)
      DO
        IF mNew = NIL THEN
          RETURN NIL 
        ELSE
          RETURN PZMatch.Pack(mNew^)
        END
      END
    END
  END MapPackedMatch;

PROCEDURE MapCandidate(
    READONLY cOld: PZCandidate.T;  (* Candidate on old curves. *)
    READONLY t0Old, t1Old: T;      (* Sampling times of old curves. *)
    READONLY t0New, t1New: T;      (* Sampling times of new curves. *)
    stride: ARRAY [0..1] OF LONG;  (* Period of "{t0Old,t1Old}" and "{t0New,t1New}". *)
    stepNew: ARRAY [0..1] OF LONG; (* Sampling step of new curves. *)
    extra: ARRAY [0..1] OF INT;    (* Extra steps to add to each end of each segment *)
  ): PZCandidate.T =
  BEGIN
    WITH
      sOld = cOld.seg,
      sNew = PZSegment.Pair{
        PZSegment.Expand(MapSegment(sOld[0], t0Old, t0New, stride[0]), extra[0], extra[0]),
        PZSegment.Expand(MapSegment(sOld[1], t1Old, t1New, stride[1]), extra[1], extra[1])
      },
      pmNew = MapPackedMatch(cOld.pm, sOld, t0Old, t1Old, sNew, t0New, t1New, stride),
      totLengthNew =
        FLOAT(sNew[0].ns - 1, LONG) * stepNew[0] +
        FLOAT(sNew[1].ns - 1, LONG) * stepNew[1]
    DO
      WITH
        mapCand = PZCandidate.T{
          seg := sNew,
          mismatch := 0.0d0,
          length := totLengthNew,
          matchedLength := 0.0d0,
          pm := pmNew
        }
      DO
        RETURN mapCand
      END;
    END;
  END MapCandidate;

BEGIN
END PZLRChain.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
