MODULE PZPlot;

(* Last edited on 1999-11-08 04:30:48 by hcgl *)

IMPORT Fmt;
IMPORT LR2;
IMPORT PZMatch, PZLR3Chain, PZSegment, PZCandidate, PSPlot;

FROM Math IMPORT exp, sin, cos;
FROM PZTypes IMPORT LONG, INT, NAT, BOOL;

FROM PSPlot IMPORT Color, Axis, Invisible;

CONST Pi = 3.1415926535897932384d0;

PROCEDURE Curve(
    f: PSPlot.File;
    READONLY c: PZLR3Chain.T;
    closed: BOOL;
    step: NAT := 1;
    phase: INT := 0;
  ) =
  VAR k: NAT;
  BEGIN
    IF step = 0 THEN step := 1 END;
    WITH m = NUMBER(c), h = phase MOD step DO
      k := 0;
      FOR  i := 1 TO m-2 DO
        IF i MOD step = h THEN
          f.segment(c[k,0], c[k,1], c[i,0], c[i,1]);
          k := i
        END
      END;
      f.segment(c[k,0], c[k,1], c[m-1,0], c[m-1,1]);
      IF closed THEN
        f.segment(c[m-1,0], c[m-1,1], c[0,0], c[0,1])
      END;
    END
  END Curve;

PROCEDURE Dots(
    f: PSPlot.File;
    READONLY c: PZLR3Chain.T;
    size: REAL := 2.0;
    step: NAT := 1;
    phase: INT := 0;
  ) =
  BEGIN
    IF step = 0 THEN step := 1 END;
    WITH m = NUMBER(c), h = phase MOD step DO
      FOR  i := 0 TO m-1 DO
        IF i MOD step = h THEN
          f.dot(c[i,0], c[i,1], size);
        END
      END;
    END
  END Dots;

PROCEDURE Match(
    f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY vA, vB : PZLR3Chain.T;
    READONLY segA, segB: PZSegment.T;
    step: NAT := 1;
    phaseA, phaseB: INT := 0;
  ) =
  VAR iA, iB: NAT;
      sAnt: NAT;
  BEGIN
    IF step = 0 THEN step := 1 END;
    <* ASSERT NUMBER(vA) = segA.tot *>
    <* ASSERT NUMBER(vB) = segB.tot *>
    WITH
      step2 = step+step,
      h = (phaseA + phaseB) MOD step2
    DO
      FOR i:= 0 TO LAST(m) DO
        IF segA.rev THEN
          iA := (segA.ini + segA.ns - 1 - m[i,0]) MOD segA.tot
        ELSE
          iA := (segA.ini + m[i,0]) MOD segA.tot
        END;
        <* ASSERT (iA - segA.ini) MOD segA.tot < segA.ns *>
        IF segB.rev THEN
          iB := (segB.ini + segB.ns - 1 - m[i,1]) MOD segB.tot
        ELSE
          iB := (segB.ini + m[i,1]) MOD segB.tot
        END;
        <* ASSERT (iB - segB.ini) MOD segB.tot < segB.ns *>
        WITH s = iA + iB DO
          IF i = 0 OR i = LAST(m)
          OR s MOD step2 = h
          OR (sAnt + 1) MOD step2 = h THEN
            f.segment(vA[iA,0], vA[iA,1], vB[iB,0], vB[iB,1])
          END;
          sAnt := s
        END
      END
    END
  END Match;

PROCEDURE Segment(
    READONLY f: PSPlot.File;
    s: PZSegment.T;
    READONLY c: PZLR3Chain.T;  (* Mapped chain *)
    closed: BOOL;              (* TRUE considers the curve closed. *)
    step: NAT := 1;            (* Plot every "step" samples. *)
    phase: INT := 0;           (* Phase of dot sampling. *)
  ) =
  BEGIN
    IF step = 0 THEN step := 1 END;
    <* ASSERT NUMBER(c) = s.tot *>
    WITH
      m = NUMBER(c),
      ini = s.ini MOD m,
      fin = (ini + s.ns - 1) MOD m
    DO
      IF s.ns > m THEN
        <* ASSERT closed *>
        Curve(f, c, FALSE, step, phase);
        f.segment(c[m-1][0], c[m-1][1], c[0][0], c[0][1]);
      ELSIF ini <= fin THEN
        Curve(f, SUBARRAY(c, ini, s.ns), FALSE, step, phase - ini);
      ELSE
        Curve(f, SUBARRAY(c, ini, m-ini), FALSE, step, phase - ini);
        IF closed THEN f.segment(c[m-1][0], c[m-1][1], c[0][0], c[0][1]) END;
        Curve(f, SUBARRAY(c, 0, fin+1), FALSE, step, phase);
      END;
    END
  END Segment;
  
PROCEDURE SegmentDots(
    READONLY f: PSPlot.File;
    s: PZSegment.T;
    READONLY c: PZLR3Chain.T;  (* Mapped chain *)
    size: REAL := 2.0;
    step: NAT := 1;            (* Plot every "step" samples. *)
    phase: INT := 0;           (* Sampling phase *)
  ) =
  BEGIN
    IF step = 0 THEN step := 1 END;
    <* ASSERT NUMBER(c) = s.tot *>
    WITH
      m = NUMBER(c),
      ini = s.ini MOD m,
      fin = (ini + s.ns - 1) MOD m
    DO
      IF s.ns > m THEN
        Dots(f, c, size, step, phase)
      ELSIF ini <= fin THEN
        Dots(f, SUBARRAY(c, ini, s.ns), size, step, phase - ini);
      ELSE
        Dots(f, SUBARRAY(c, ini, m-ini), size, step, phase-ini);
        Dots(f, SUBARRAY(c, 0, fin+1), size, step, phase);
      END;
    END
  END SegmentDots;

PROCEDURE SegmentLabel(
    READONLY f: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
    label: TEXT;
    size: REAL := 12.0;
    font: TEXT := "Courier";
  ) =
  BEGIN
    WITH
      m = NUMBER(c),
      mid = (s.ini + s.ns DIV 2) MOD m,
      opp = (mid + m DIV 2) MOD m,
      midW = FLOAT(m + 2*MAX(0, m - s.ns), LONG),
      oppW = FLOAT(m, LONG)
    DO
      <* ASSERT m = s.tot *>
      f.label(
        label,
        (midW * c[mid][0] + oppW * c[opp][0])/(midW + oppW),
        (midW * c[mid][1] + oppW * c[opp][1])/(midW + oppW),
        0.5, 0.5,
        font := font,
        size := size
      );
    END;
  END SegmentLabel;

PROCEDURE SegmentPointer(
    READONLY f: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
    frac: LONG;
    width: REAL := 4.0; 
    length: REAL := 8.0; 
    inside: BOOL := TRUE;
  ) =
  VAR p: LR2.T;
  BEGIN
    WITH
      m = NUMBER(c),
      
      (* Point to point at: *)
      q = InterpolateSegment(frac, s, c),
      
      iu = (s.ini + s.ns DIV 2) MOD m,
      u = LR2.T{c[iu][0], c[iu][1]},
      iv = (iu + m DIV 2) MOD m,
      v = LR2.T{c[iv][0], c[iv][1]}
    DO
      (* Approximate center: *)
      p := LR2.Mix(0.5d0, u, 0.5d0, v);
      IF p[0] = q[0] AND p[1] = q[1] THEN
        p := LR2.T{q[0] - 1.0d0, q[1] - 1.0d0}
      END;
      IF NOT inside THEN 
        p := LR2.Mix(-1.0d0, p, 2.0d0, q)
      END;
      f.arrowhead(p[0], p[1], q[0], q[1], width, length, 1.0d0)
    END;
  END SegmentPointer;
  
PROCEDURE SegmentBrackets(
    READONLY f: PSPlot.File;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
    width: REAL := 4.0; 
    length: REAL := 8.0; 
  ) =
  VAR u, v: LR2.T;
  BEGIN
    <* ASSERT s.tot = NUMBER(c) *>
    WITH
      pIni = c[(s.ini + 1) MOD s.tot],
      qIni = c[s.ini MOD s.tot],
      pFin = c[(s.ini + s.ns - 2) MOD s.tot],
      qFin = c[(s.ini + s.ns - 1) MOD s.tot],
      p = ARRAY [0..1] OF LR2.T{LR2.T{pIni[0],pIni[1]}, LR2.T{pFin[0],pFin[1]}},
      q = ARRAY [0..1] OF LR2.T{LR2.T{qIni[0],qIni[1]}, LR2.T{qFin[0],qFin[1]}},
      d = ARRAY [0..1] OF LR2.T{LR2.T{-1.0d-3,0.0d0}, LR2.T{+1.0d-3,0.0d0}}
    DO
      FOR j := 0 TO 1 DO 
        u := p[j]; v := q[j];
        IF LR2.Dist(u,v) < 0.0001d0 THEN u := q[1-j] END;
        IF LR2.Dist(u,v) < 0.0001d0 THEN u := LR2.Add(v, d[j]) END;
        f.arrowhead(u[0], u[1], v[0], v[1], width, length, 1.0d0)
      END
    END;
  END SegmentBrackets;
  
PROCEDURE InterpolateSegment(
    frac: LONG;
    READONLY s: PZSegment.T;
    READONLY c: PZLR3Chain.T;
  ): LR2.T =
  VAR ia, ib: NAT;
  BEGIN
    WITH
      m = NUMBER(c),
      fx = frac*FLOAT(s.ns, LONG),
      ix = FLOOR(fx),
      sx = fx - FLOAT(ix, LONG)
    DO
      <* ASSERT m = s.tot *>
      IF s.rev THEN
        ia := (s.ini + s.ns - 1 - ix) MOD m; 
        ib := (s.ini + s.ns - 1 - ix - 1) MOD m;
      ELSE
        ia := (s.ini + ix) MOD m; 
        ib := (s.ini + ix + 1) MOD m;
      END;
      WITH
        a = LR2.T{c[ia][0], c[ia][1]},
        b = LR2.T{c[ib][0], c[ib][1]}
      DO
        RETURN LR2.Mix(1.0d0 - sx, a, sx, b)
      END
    END
  END InterpolateSegment;

PROCEDURE GaussianKernel(
    f: PSPlot.File;
    lambda: LONG;
    x, y: LONG;
    width: LONG;
    height: LONG;
    lineWidth: REAL;
    dotStep: LONG := 0.0d0;
    dotSize: REAL := 0.0;
  ) =

  VAR a: LR2.T;  (* Previous point on graph *)

  PROCEDURE GraphPoint(i: INT; step: LONG): LR2.T =
    BEGIN
      WITH
        zp = FLOAT(i, LONG)*step,
        xp = x + zp,
        tp = zp/lambda,
        yp = y + height * exp(-tp*tp/2.0d0)
      DO
        RETURN LR2.T{xp, yp}
      END
    END GraphPoint;

  BEGIN
    IF lineWidth > 0.0 THEN
      f.setLineWidth(lineWidth);
      WITH
        plotStep = lambda/12.0d0,
        nSteps = FLOOR((width/2.0d0) / plotStep)
      DO
        (* Draw the graph: *)
        FOR i := -nSteps TO +nSteps DO
          WITH p = GraphPoint(i, plotStep) DO
            IF i > -nSteps THEN f.segment(a[0], a[1], p[0], p[1]) END;
            a := p;
          END
        END
      END
    END;

    IF dotStep > 0.0d0 AND dotSize > 0.0 THEN
      (* Draw the dots: *)
      WITH
        nSteps = FLOOR((width/2.0d0) / dotStep)
      DO
        FOR i := -nSteps TO +nSteps DO
          WITH p = GraphPoint(i, dotStep) DO
            f.dot(p[0], p[1], dotSize)
          END
        END
      END
    END
  END GaussianKernel;

PROCEDURE CriticalCircle(
    f: PSPlot.File;
    radius: LONG;
    x, y: LONG;
    width: LONG;
    height: LONG;
    lineWidth: REAL;
    dotStep: LONG := 0.0d0;
    dotSize: REAL := 0.0;
  ) =
  
  VAR
    xc, yc: LONG; (* Circle center *)
    a: LR2.T;     (* Previous point on graph *)

  PROCEDURE CirclePoint(i: INT; step: LONG): LR2.T =
    BEGIN
      WITH
        theta = FLOAT(i, LONG)*step/radius,
        xp = xc + radius*sin(theta),
        yp = yc + radius*cos(theta)
      DO
        RETURN LR2.T{xp, yp}
      END
    END CirclePoint;

  PROCEDURE Inside(READONLY p: LR2.T): BOOL =
    BEGIN
      RETURN 
        ABS(p[0] - x) <= width/2.0d0 AND 
        p[1] >= y AND
        p[1] <= y + height
    END Inside;

  BEGIN
    WITH halfHeight = height/2.0d0 DO
      (* Assumes width >= height; will not look good if width < height. *)
      xc := x;
      IF radius <= halfHeight THEN
        yc := y + halfHeight 
      ELSE
        yc := y + height - radius
      END;
    END;
    IF lineWidth > 0.0 THEN
      f.setLineWidth(lineWidth);
      WITH
        nSteps = 32,
        plotStep = Pi*radius/FLOAT(nSteps, LONG)
      DO
        (* Draw the graph: *)
        FOR i := -nSteps TO +nSteps DO
          WITH p = CirclePoint(i, plotStep) DO
            IF i > -nSteps AND Inside(a) AND Inside(p) THEN
              f.segment(a[0], a[1], p[0], p[1])
            END;
            a := p;
          END
        END
      END
    END;

    IF dotStep > 0.0d0 AND dotSize > 0.0 THEN
      (* Draw the dots: *)
      WITH
        nSteps = FLOOR(Pi*radius / dotStep)
      DO
        FOR i := -nSteps TO +nSteps DO
          WITH p = CirclePoint(i, dotStep) DO
            f.dot(p[0], p[1], dotSize)
          END
        END
      END
    END
  END CriticalCircle;

PROCEDURE ScaleBar(
    f: PSPlot.File;
    lambda: LONG;
    count: NAT;
    x, y: LONG;
    lineWidth: REAL;
    tickLength: REAL;
  ) =
  BEGIN
    IF lineWidth > 0.0 THEN
      f.setLineWidth(lineWidth);
      WITH
        tickExtent = FLOAT(tickLength/2.0, LONG) / f.getScaleFactor(PSPlot.Axis.Y),
        halfWidth = FLOAT(count, LONG)*lambda/2.0d0
      DO
        f.segment(x - halfWidth, y, x + halfWidth, y);
        FOR i := 0 TO count DO
          WITH xTick = x - halfWidth + FLOAT(i, LONG)*lambda DO
            f.segment(xTick, y - tickExtent, xTick, y + tickExtent)
          END
        END
      END
    END
  END ScaleBar;

PROCEDURE Candidate(
    f: PSPlot.File;
    READONLY cand: PZCandidate.T;
    READONLY c0, c1: PZLR3Chain.T;  (* Chains, already mapped; *)
    whole: BOOL;
    closed: BOOL;
    dots: BOOL;
    pairing: BOOL;
    labels: BOOL;
    pointers: BOOL;
    step: NAT := 1;
  ) =
  VAR m: REF PZMatch.T;
  BEGIN
    f.setLineSolid();
    f.setFillColor(PSPlot.Invisible);
    WITH 
      s0 = cand.seg[0], mid0 = s0.ini + s0.ns DIV 2,
      s1 = cand.seg[1], mid1 = s1.ini + s1.ns DIV 2
    DO
      IF pairing THEN
        f.setLineWidth(0.1);
        f.setLineColor(Color{0.50, 0.50, 0.50});
        IF cand.pm # NIL THEN
          m := PZMatch.Unpack(cand.pm^)
        ELSE
          m := PZMatch.MostPerfect(s0.ns, s1.ns)
        END;
        Match(f, m^, c0, c1, s0, s1, step, mid0, mid1)
      END;
      HalfCandidate(f, s0, c0, PSPlot.Red,
        whole := whole, closed := closed, 
        dots := dots, label := labels, pointers := pointers,
        step := step
      );
      HalfCandidate(f, s1, c1, PSPlot.Blue,
        whole := whole, closed := closed, 
        dots := dots, label := labels, pointers := pointers,
        step := step
      )
    END;
  END Candidate;

PROCEDURE HalfCandidate(
    f: PSPlot.File;
    s: PZSegment.T;
    READONLY c: PZLR3Chain.T;  (* Chain, already mapped; *)
    color: PSPlot.Color;
    whole: BOOL;
    closed: BOOL;
    dots: BOOL;
    label: BOOL;
    pointers: BOOL;
    step: NAT := 1;
  ) =
  BEGIN
    <* ASSERT s.tot = NUMBER(c) *>
    IF label THEN
      f.setTextColor(PSPlot.Black);
      SegmentLabel(f, s, c, Fmt.Int(s.cvx));
    END;
    WITH phase = s.ini + s.ns DIV 2 DO
      IF whole THEN
        f.setLineWidth(0.1);
        f.setLineColor(PSPlot.Black);
        WITH r = PZSegment.Complement(s) DO
          Segment(f, r, c, closed, step, phase);
          IF dots THEN SegmentDots(f, r, c, 2.0, step, phase) END;
        END;
      END;
      f.setLineWidth(0.2);
      f.setLineColor(color);
      Segment(f, s, c, closed, step, phase);
      IF dots THEN SegmentDots(f, s, c, 2.0, step, phase) END;
      IF pointers THEN 
        f.setLineWidth(0.1);
        SegmentPointer(f, s, c, 0.0d0, 1.0/0.1, 2.0/0.1, inside := TRUE);
        SegmentPointer(f, s, c, 1.0d0, 1.0/0.1, 2.0/0.1, inside := TRUE);
      END
    END
  END HalfCandidate;

PROCEDURE ScaleIcon(
    f: PSPlot.File;
    filterLambda: LONG;
    filterDotStep: LONG;
    style: ScaleIconStyle
  ) =
  VAR factor: LONG;  (* The icon will be "2*factor*filterLambda" wide. *) 
  CONST CriticalAlpha = 0.8577638846d0;  (* sqrt(2/e) *)
  BEGIN
    CASE style OF
    | ScaleIconStyle.Kernel => factor := 2.5d0;
    | ScaleIconStyle.Circle => factor := 1.0d0/CriticalAlpha;
    END;
    WITH
      xRange = f.getRange(Axis.X),
      yRange = f.getRange(Axis.Y),
      tickLength = 1.0,         (* In millimeters *)
      (* Leave 2mm of blank space around drawing: *)
      xMargin = 2.0d0 / f.getScaleFactor(Axis.X),
      yMargin = 2.0d0 / f.getScaleFactor(Axis.Y),
      (* Leave another 2mm of distance from the drawing's border: *)
      xSafety =  2.0d0 / f.getScaleFactor(Axis.X),
      ySafety =  2.0d0 / f.getScaleFactor(Axis.Y),
      (* Make the icon window 8mm tall: *)
      iconHeight = 8.0d0 / f.getScaleFactor(Axis.Y),
      (* The tick marks of the scale bar will extend below "y": *)
      barHeight = FLOAT(tickLength/2.0, LONG) / f.getScaleFactor(Axis.Y),
      barDepth = barHeight,
      iconDepth = barDepth,
      (* Ideally the icon window shoudl contain the bar AND the graph/circle: *)
      idealIconWidth = 2.0d0*MAX(1.0d0, factor)*filterLambda,
      (* At the very least, it must contain the scale bar: *)
      minIconWidth = 2.0d0*filterLambda,
      (* On the other hand, it should not be too wide: *)
      maxIconWidth = (xRange.max - xRange.min - 2.0d0*(xSafety+xMargin))/3.0d0,
      (* In conclusion: *)
      iconWidth = MIN(maxIconWidth, idealIconWidth),
      (* Compute midpoint of bottom side of window: *)
      x = xRange.max - xSafety - xMargin - iconWidth/2.0d0,
      y = yRange.min + ySafety + yMargin + iconDepth
    DO
      IF iconWidth >= minIconWidth  THEN
        (* Blank out the background: *)
        f.setLineSolid();
        f.setLineColor(Color{0.0, 0.0, 0.0});
        f.setFillColor(Color{1.0, 1.0, 1.0});
        WITH dx = iconWidth/2.0d0 + xMargin DO
          f.rectangle(x - dx, x + dx, y - yMargin - iconDepth, y + yMargin + iconHeight)
        END;
        (* Plot the kernel or circle, with lines: *)
        f.setLineColor(Color{0.0, 0.0, 0.5});
        f.setFillColor(Invisible);
        f.setLineWidth(0.15);
        CASE style OF
        | ScaleIconStyle.Kernel => 
            GaussianKernel(f, 
              lambda := filterLambda,
              x := x, y := y,
              width := iconWidth, height := iconHeight,
              lineWidth := 0.15,
              dotStep := filterDotStep, dotSize := 2.0
            );
        | ScaleIconStyle.Circle => 
            CriticalCircle(f, 
              radius := factor * filterLambda,
              x := x, y := y + barHeight + yMargin,
              width := iconWidth, height := iconHeight - barHeight - yMargin,
              lineWidth := 0.15,
              dotStep := filterDotStep, dotSize := 2.0
            );
        END;
        f.setLineSolid();
        f.setLineColor(Color{0.0, 0.0, 0.5});
        f.setFillColor(Invisible);
        ScaleBar(f, lambda := filterLambda, count := 2,
          x := x, y := y, lineWidth := 0.15, tickLength := tickLength
        );
      END;
    END
  END ScaleIcon;

PROCEDURE SegSegGrid(
    READONLY f: PSPlot.File;
    READONLY s: SegmentPair;
    READONLY sRef: SegmentPair;
    cellSize: LONG;
    drawDiags: BOOL;
  ) =
  BEGIN
    WITH
      n0 = s[0].ns,
      lo0 = (FLOAT(s[0].ini - sRef[0].ini, LONG) + 0.5d0) * cellSize,
      hi0 = lo0 + FLOAT(n0 - 1, LONG) * cellSize,

      n1 = s[1].ns,
      lo1 = (FLOAT(s[1].ini - sRef[1].ini, LONG) + 0.5d0) * cellSize,
      hi1 = lo1 + FLOAT(n1 - 1, LONG) * cellSize
    DO
      FOR i0 := 0 TO n0-1 DO
        WITH xt0 = lo0 + FLOAT(i0, LONG) * cellSize DO
          f.segment(xt0, lo1, xt0, hi1)
        END
      END;
      FOR i1 := 0 TO n1 - 1 DO
        WITH xt1 = lo1 + FLOAT(i1, LONG) * cellSize DO
          f.segment(lo0, xt1, hi0, xt1)
        END
      END;
      IF drawDiags THEN
        FOR k := 2 TO n0 + n1 - 2 DO
          WITH
            da0 = FLOAT(n0 - MIN(k, n0), LONG) * cellSize,
            da1 = FLOAT(MAX(k, n0) - n0, LONG) * cellSize,

            db0 = FLOAT(n0 + n1 - MAX(k, n1) - 1, LONG) * cellSize,
            db1 = FLOAT(MIN(k, n1) - 1, LONG) * cellSize
          DO
            IF s[0].rev = s[1].rev THEN
              f.segment(lo0 + da0, lo1 + da1, lo0 + db0, lo1 + db1)
            ELSE
              f.segment(hi0 - da0, lo1 + da1, hi0 - db0, lo1 + db1)
            END
          END
        END
      END
    END
  END SegSegGrid;

TYPE IntPair = ARRAY [0..1] OF INT;

CONST BadPair = IntPair{LAST(INT), LAST(INT)};

PROCEDURE ReducePair(
    p: PZMatch.Pair;
    READONLY s, sRef: SegmentPair;
    dLo, dHi: INT;
  ): IntPair =
  (*
    Given an index pair relative to segments "s[0]" and "s[1]",
    returns the corresponding index pair relative to segments "sRef[0]"
    and "sRef[1]".   Does NOT consider that the chains are periodic.
    Returns "BadPair" if either component "j" of the result does not
    lie in the range "[dLo .. sRef[j].ns + dHi]". *)
  VAR r: IntPair;
  BEGIN
    FOR j := 0 TO 1 DO
      <* ASSERT s[j].cvx = sRef[j].cvx *>
      <* ASSERT s[j].tot = sRef[j].tot *>
      WITH
        i = PZSegment.RelIndex(PZSegment.AbsIndex(p[j], s[j]), sRef[j])
      DO
        IF i >= dLo AND i <= sRef[j].ns + dHi THEN
          r[j] := i
        ELSE
          RETURN BadPair
        END
      END
    END;
    RETURN r
  END ReducePair;

PROCEDURE Coord(iRel: INT; READONLY sRef: PZSegment.T; step: LONG): LONG =
  (*
    Given a sample index "iRel" relative to a segment "sRef",
    returns the corresponding plot coordinate, assuming that
    the minimum coordinate in "sRef" is "0.5d0 * step". *)
  BEGIN
    IF sRef.rev THEN
      RETURN (FLOAT(sRef.ns - 1 - iRel, LONG) + 0.5d0) * step
    ELSE
      RETURN (FLOAT(iRel, LONG) + 0.5d0) * step
    END
  END Coord;

PROCEDURE MatchBoxes(
    READONLY f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY s: SegmentPair;
    READONLY sRef: SegmentPair;
    cellSize: LONG;
    boxSize: REAL;
  ) =
  BEGIN
    WITH
      r = (FLOAT(boxSize, LONG)) / 2.0d0,
      extra = CEILING(r/cellSize + 0.1d0 * cellSize)
    DO
      FOR k := 0 TO LAST(m) DO
        WITH p = ReducePair(m[k], s, sRef, -extra, +extra) DO
          IF p # BadPair THEN
            WITH
              xp = Coord(p[0], sRef[0], cellSize),
              yp = Coord(p[1], sRef[1], cellSize)
            DO
              f.rectangle(xp - r, xp + r, yp - r, yp + r)
            END
          END
        END
      END;
    END
  END MatchBoxes;

PROCEDURE MatchPath(
    READONLY f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY s: SegmentPair;
    READONLY sRef: SegmentPair;
    cellSize: LONG;
  ) =
  BEGIN
    FOR k := 1 TO LAST(m) DO
      WITH
        p = ReducePair(m[k-1], s, sRef, -1, +1),
        q = ReducePair(m[k],   s, sRef, -1, +1)
      DO
        IF p # BadPair AND q # BadPair THEN
          WITH
            xp = Coord(p[0], sRef[0], cellSize),
            yp = Coord(p[1], sRef[1], cellSize),

            xq = Coord(q[0], sRef[0], cellSize),
            yq = Coord(q[1], sRef[1], cellSize)
          DO
            f.segment(xp, yp, xq, yq)
          END
        END
      END
    END;
  END MatchPath;

PROCEDURE MatchDots(
    READONLY f: PSPlot.File;
    READONLY m: PZMatch.T;
    READONLY s: SegmentPair;
    READONLY sRef: SegmentPair;
    cellSize: LONG;
    dotSize: REAL;
  ) =
  BEGIN
    IF dotSize <= 0.0 THEN RETURN END;
    f.setLineWidth(1.0);
    WITH
      r = (FLOAT(dotSize, LONG)) / 2.0d0,
      extra = CEILING(FLOAT(r, LONG)/cellSize + 0.1d0 * cellSize)
    DO
      FOR k := 0 TO LAST(m) DO
        WITH p = ReducePair(m[k], s, sRef, -extra, +extra) DO
          IF p # BadPair THEN
            WITH
              xp = Coord(p[0], sRef[0], cellSize),
              yp = Coord(p[1], sRef[1], cellSize)
            DO
              f.dot(xp, yp, dotSize)
            END
          END
        END
      END;
    END
  END MatchDots;

PROCEDURE TryAllShifts(
    READONLY s: SegmentPair;
    READONLY sRef: SegmentPair;
    proc: SegSegPlotProc;
  ) =

  PROCEDURE Overlap(sa, sb: PZSegment.T): BOOL =
    (*
      TRUE if "sa" intersects "sb", ignoring wrap-around. *)
    BEGIN
      RETURN sa.ini < sb.ini + sb.ns AND sb.ini < sa.ini + sa.ns
    END Overlap;

  VAR
    t: SegmentPair;     (* Translated copy of "s" *)
    tRef: SegmentPair;  (* Reduced copy of "sRef" *)

  BEGIN
    (* Get a copy "tRef" of "sRef" with "ini" in "[tot..2*tot-1]": *)
    tRef := sRef;
    FOR j := 0 TO 1 DO
      <* ASSERT sRef[j].tot = s[j].tot *>
      tRef[j].ini := (sRef[j].ini MOD sRef[j].tot) + sRef[j].tot;
    END;
    (* Find all translates "t" of "s" that overlap "tRef": *)
    t := s;
    FOR k0 := 0 TO 2 DO
      FOR k1 := 0 TO 2 DO
        t[0].ini := s[0].ini MOD s[0].tot + k0*s[0].tot;
        t[1].ini := s[1].ini MOD s[1].tot + k1*s[1].tot;
        IF Overlap(t[0], tRef[0]) AND Overlap(t[1], tRef[1]) THEN
          proc(t, tRef)
        END
      END
    END
  END TryAllShifts;

BEGIN
END PZPlot.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
