MODULE PZProc;

(* Last edited on 1999-09-03 02:54:17 by hcgl *)

FROM Math IMPORT erf, sqrt, log, exp;
FROM PZTypes IMPORT NAT, LONG, INT, BOOLS;

PROCEDURE ReduceToPeriod(VAR (*IO*) t: LONG; tLo, tHi: LONG; VAR (*OUT*) k: INT) =
  BEGIN
    IF t >= tLo AND t < tHi THEN 
      k := 0
    ELSE
      WITH T = tHi - tLo DO
        <* ASSERT T > 0.0d0 *>
        k := FLOOR((t - tLo)/T);
        t := t - FLOAT(k,LONG)*T;
        WHILE t < tLo DO t := t + T; INC(k) END;
        WHILE t >= tHi DO t := t - T; DEC(k) END;
      END
    END
  END ReduceToPeriod;

CONST
  SqrtTwo = 1.4142135623730950488d0;

PROCEDURE Squeeze(x: LONG; sigma: LONG): LONG = 
  BEGIN
    RETURN erf(x/(SqrtTwo*sigma))
  END Squeeze;

PROCEDURE Stretch(x: LONG; sigma: LONG): LONG =
  VAR a, b, m: LONG;
  BEGIN
    IF x <= -1.0d0 THEN RETURN FIRST(LONG) END;
    IF x >= +1.0d0 THEN RETURN LAST(LONG)  END;
    a := -1.0d0; b := +1.0d0;
    WHILE x < Squeeze(a, sigma) DO
      VAR t := a - 2.0d0*(b - a); BEGIN b := a; a := t END
    END;
    WHILE x > Squeeze(b, sigma) DO
      VAR t := b + 2.0d0*(b - a); BEGIN a := b; b := t END
    END;
    WHILE b - a > 0.0001d0 DO
      m := (a + b)/ 2.0d0;
      IF x > Squeeze(b, sigma) THEN a := m ELSE b := m END
    END;
    RETURN (a + b)/ 2.0d0;
  END Stretch;

PROCEDURE OldSqueeze(x: LONG; epsilon, delta: LONG): LONG =
  BEGIN
    WITH
      r = delta*x/epsilon
    DO
      RETURN log(r + sqrt(r*r + 1.0d0))/delta
    END
  END OldSqueeze;

PROCEDURE OldStretch(x: LONG; epsilon, delta: LONG): LONG =
  BEGIN
    WITH
      h = exp(x * delta),
      r = (h - 1.0d0)/(2.0d0*h)
    DO
      RETURN epsilon*r/delta
    END
  END OldStretch;

PROCEDURE Interpolate(t: LONG; t0, p0: LONG; t1, p1 : LONG) : LONG =
  VAR p: LONG;
  BEGIN 
    IF (t1 - t0) = 0.0d0 THEN
      WITH 
        alfa = (t - t0) / (t1 - t0)
      DO 
        p := alfa * p1 + (1.00d0 - alfa) * p0;
      END
    ELSE
      p := 0.5d0 * p1 + 0.5d0* p0;
    END;
    RETURN p;
  END Interpolate;

PROCEDURE CubicInterpolate(
    t: LONG; 
    t0, p0: LONG; 
    t1, p1: LONG;
    t2, p2: LONG;
    t3, p3: LONG;
  ): LONG =
  BEGIN 
    <* ASSERT t0 <= t1 AND t1 <= t2 AND t2 <= t3 *>
    (* Test for degenerate intervals: *)
    WITH eps = FLOAT(1.0e-6,LONG)*(t3 - t0) DO
      IF eps = 0.0d0 OR (t2-t1) < eps THEN
        (* Middle interval is degenerate: *)
        RETURN 0.5d0*p1 + 0.5d0*p2
      ELSIF (t1-t0) < eps OR (t3-t2) < eps THEN
        (* End intervals degenerate: use linear interpolation *)
        RETURN Interpolate(t, t1, p1, t2, p2)
      END
    END;
      
    WITH
      p01 = Interpolate(t, t0, p0, t1, p1),
      p12 = Interpolate(t, t1, p1, t2, p2),
      p23 = Interpolate(t, t2, p2, t3, p3),
      p012 = Interpolate(t, t0, p01, t2, p12),
      p123 = Interpolate(t, t1, p12, t3, p23),
      p0123 = Interpolate(t, t0, p012, t3, p123)
    DO
      RETURN p0123
    END
  END CubicInterpolate;

PROCEDURE SelectAll(n: NAT): REF BOOLS =
  BEGIN
    WITH bR = NEW(REF BOOLS, n), b = bR^ DO
      FOR i := 0 TO n-1 DO b[i] := TRUE END;
      RETURN bR
    END
  END SelectAll;

BEGIN 
END PZProc.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
